# Based on squidfunk/mkdocs-material badges
# https://github.com/squidfunk/mkdocs-material/blob/master/material/overrides/hooks/shortcodes.py


from __future__ import annotations

import re
import typing as t
from re import Match

VERSION_ICON = "material-tag-outline"


def on_page_markdown(markdown: str, **kwargs: t.Any):
    return re.sub(r"<!-- md:(\w+)(.*?) -->", _gen_badge, markdown, flags=re.I | re.M)


def _gen_badge(match: Match[t.Any]):
    name, arg_ls = match.groups()
    args: list[str] = [a for a in arg_ls.strip().split(" ")]
    if name == "version":
        return _version_badge(args[0])

    raise Exception(f"Unknown name for helper badge: {name}")


def _badge(icon: str, text: str = "", icon_tooltip: str | None = None):
    tooltip = f'{{ title="{icon_tooltip}" }}' if icon_tooltip else ""
    return "".join(
        [
            '<span class="clypi-badge">',
            *(
                [f'<span class="clypi-badge-icon">:{icon}:{tooltip}</span>']
                if icon
                else []
            ),
            *([f'<span class="clypi-badge-text">{text}</span>'] if text else []),
            "</span>",
        ]
    )


def _version_badge(version: str):
    return _badge(icon=VERSION_ICON, text=f"{version}", icon_tooltip="Minimum version")



import inspect
import typing as t
from enum import Enum
from types import NoneType, UnionType

P = t.ParamSpec("P")
R = t.TypeVar("R")


def ignore_annotated(fun: t.Callable[P, R]) -> t.Callable[P, R]:
    def inner(*args: P.args, **kwargs: P.kwargs) -> R:
        if t.get_origin(args[0]) == t.Annotated:
            args_ls = list(args)
            args_ls[0] = t.get_args(args[0])[0]
            args = tuple(args_ls)  # type: ignore
        return fun(*args, **kwargs)

    return inner


@ignore_annotated
def is_list(_type: t.Any) -> t.TypeGuard[list[t.Any]]:
    return t.get_origin(_type) in (list, t.Sequence)


@ignore_annotated
def list_inner(_type: t.Any) -> t.Any:
    return t.get_args(_type)[0]


@ignore_annotated
def is_tuple(_type: t.Any) -> t.TypeGuard[tuple[t.Any]]:
    return t.get_origin(_type) is tuple


@ignore_annotated
def tuple_inner(_type: t.Any) -> tuple[list[t.Any], int | None]:
    """
    Returns the list of types for the tuple and how many items
    it accepts
    """
    # TODO: can be made more efficient
    inner_types = list(t.get_args(_type))
    if inner_types[-1] is Ellipsis:
        return [inner_types[0]], None
    return inner_types, len(inner_types)


@ignore_annotated
def is_union(_type: t.Any) -> t.TypeGuard[UnionType]:
    return isinstance(_type, UnionType)


@ignore_annotated
def union_inner(_type: t.Any) -> list[t.Any]:
    return list(t.get_args(_type))


@ignore_annotated
def is_literal(_type: t.Any) -> bool:
    return t.get_origin(_type) == t.Literal


@ignore_annotated
def literal_inner(_type: t.Any) -> list[t.Any]:
    return list(t.get_args(_type))


@ignore_annotated
def tuple_size(_type: t.Any) -> float:
    args = _type.__args__
    if args[-1] is Ellipsis:
        return float("inf")
    return len(args)


@ignore_annotated
def is_none(_type: t.Any) -> t.TypeGuard[type[None]]:
    return _type is NoneType


@ignore_annotated
def is_enum(_type: t.Any) -> t.TypeGuard[type[Enum]]:
    return inspect.isclass(_type) and issubclass(_type, Enum)


@ignore_annotated
def has_metavar(_type: t.Any) -> bool:
    return is_enum(_type) or is_literal(_type)


@ignore_annotated
def remove_optionality(_type: t.Any) -> t.Any:
    if not isinstance(_type, UnionType):
        return _type

    new_args: list[t.Any] = []
    for arg in _type.__args__:
        if arg is not type[None]:
            new_args.append(arg)

    if len(new_args) == 1:
        return new_args[0]

    return t.Union[*new_args]



import typing as t

from clypi._colors import ColorType, Styler
from clypi._components.align import AlignType
from clypi._components.align import align as _align
from clypi._components.wraps import wrap
from clypi._data.boxes import Boxes as _Boxes
from clypi._util import get_term_width, visible_width

Boxes = _Boxes


T = t.TypeVar("T", bound=list[str] | str)


def boxed(
    lines: T,
    width: t.Literal["auto", "max"] | int = "auto",
    style: Boxes = Boxes.HEAVY,
    align: AlignType = "left",
    title: str | None = None,
    color: ColorType | None = None,
) -> T:
    box = style.value
    c = Styler(fg=color)

    def _iter_box(
        lines: t.Iterable[str],
        width: int,
    ):
        # Top bar
        nonlocal title
        top_bar_width = width - 3
        if title:
            top_bar_width = width - 5 - visible_width(title)
            title = f" {title} "
        else:
            title = ""
        yield c(box.tl + box.x + title + box.x * top_bar_width + box.tr)

        # Body
        for line in lines:
            # Remove two on each side due to the box edge and padding
            max_text_width = -2 + width - 2

            # Wrap it in case each line is longer than expected
            wrapped = wrap(line, max_text_width)
            for sub_line in wrapped:
                aligned = _align(sub_line, align, max_text_width)
                yield c(box.y) + " " + aligned + " " + c(box.y)

        # Footer
        yield c(box.bl + box.x * (width - 2) + box.br)

    def _get_width(lines: list[str]):
        if isinstance(width, int):
            return width

        if width == "max":
            return get_term_width()

        # Width is auto
        max_visible_width = max(visible_width(line) for line in lines)
        # Add two on each side for the box edge and padding
        return 2 + max_visible_width + 2

    if isinstance(lines, list):
        computed_width = _get_width(lines)
        return t.cast(T, list(_iter_box(lines, width=computed_width)))

    act_lines = lines.split("\n")
    computed_width = _get_width(act_lines)
    return t.cast(T, "\n".join(_iter_box(act_lines, width=computed_width)))



import pytest

from clypi import OverflowStyle, wrap


@pytest.mark.parametrize(
    "s,overflow_style,expected",
    [
        ("a", "ellipsis", ["a"]),
        ("a" * 20, "ellipsis", ["a" * 20]),
        ("a" * 21, "ellipsis", ["a" * 19 + "…"]),
        ("a", "wrap", ["a"]),
        ("a" * 20, "wrap", ["a" * 20]),
        ("a" * 21, "wrap", ["a" * 20, "a"]),
        ("a" * 45, "wrap", ["a" * 20, "a" * 20, "a" * 5]),
    ],
)
def test_wrapping(s: str, overflow_style: OverflowStyle, expected: str):
    width = 20
    result = wrap(s, width, overflow_style)
    assert result == expected
    assert len(result) <= width



from __future__ import annotations

import typing as t
from collections import defaultdict
from dataclasses import dataclass
from functools import cached_property

from clypi import _type_util
from clypi._cli.arg_parser import dash_to_snake
from clypi._colors import ColorType, style
from clypi._components.boxed import boxed
from clypi._components.indented import indented
from clypi._components.stack import stack
from clypi._exceptions import format_traceback

if t.TYPE_CHECKING:
    from clypi import Command
    from clypi._cli.arg_config import Config


class Formatter(t.Protocol):
    def format_help(
        self,
        full_command: list[str],
        description: str | None,
        epilog: str | None,
        options: list[Config[t.Any]],
        positionals: list[Config[t.Any]],
        subcommands: list[type[Command]],
        exception: Exception | None,
    ) -> str: ...


@dataclass
class ClypiFormatter:
    boxed: bool = True
    show_option_types: bool = False
    show_inherited_options: bool = True
    normalize_dots: t.Literal[".", ""] | None = ""

    @cached_property
    def theme(self):
        from clypi._configuration import get_config

        return get_config().theme

    def _maybe_norm_help(self, message: str) -> str:
        """
        Utility function to add or remove dots from the end of all option/arg
        descriptions to have a more consistent formatting experience.
        """
        message = message.rstrip()
        if message and self.normalize_dots == "." and message[-1].isalnum():
            return message + "."
        if message and self.normalize_dots == "" and message[-1] == ".":
            return message[:-1]
        return message

    def _maybe_boxed(
        self, *columns: list[str], title: str, color: ColorType | None = None
    ) -> str:
        first_col, *rest = columns

        # Filter out empty columns
        rest = list(filter(any, rest))

        if not self.boxed:
            section_title = self.theme.section_title(title)

            # For non-boxed design, we just indent the first col a bit so that it looks
            # like it's inside the section
            stacked = stack(indented(first_col), *rest)
            return f"{section_title}\n{stacked}"

        stacked = stack(first_col, *rest, lines=True)
        return "\n".join(boxed(stacked, width="max", title=title, color=color))

    def _format_option_value(self, option: Config[t.Any]):
        if option.nargs == 0:
            return ""
        placeholder = dash_to_snake(option.name).upper()
        return self.theme.placeholder(f"<{placeholder}>")

    def _format_option(self, option: Config[t.Any]) -> tuple[str, ...]:
        help = self._maybe_norm_help(option.help or "")

        # E.g.: -r, --requirements <REQUIREMENTS>
        name = self.theme.long_option(option.display_name)
        short_usage = (
            self.theme.short_option(option.short_display_name) if option.short else ""
        )
        usage = name
        if short_usage:
            usage = short_usage + ", " + usage
        if not self.show_option_types:
            usage += " " + self._format_option_value(option)

        # E.g.: TEXT
        type_str = ""
        type_upper = str(option.parser).upper()
        if self.show_option_types:
            type_str = self.theme.type_str(type_upper)
        elif _type_util.has_metavar(option.arg_type):
            help = help + " " + type_upper if help else type_upper

        return usage, type_str, help

    def _format_option_group(
        self, title: str, options: list[Config[t.Any]]
    ) -> str | None:
        if not options:
            return None

        usage: list[str] = []
        type_str: list[str] = []
        help: list[str] = []
        for o in options:
            # Hidden options do not get displayed for the user
            if o.hidden:
                continue

            u, ts, hp = self._format_option(o)
            usage.append(u)
            type_str.append(ts)
            help.append(hp)

        return self._maybe_boxed(usage, type_str, help, title=title)

    def _format_options(self, options: list[Config[t.Any]]) -> str | None:
        if not options:
            return None

        groups: dict[str | None, list[Config[t.Any]]] = defaultdict(list)

        # We set an empty group first so that non-group options always render first
        groups[None] = []

        # Group by option group
        for o in options:
            if o.inherited and not self.show_inherited_options:
                continue
            groups[o.group].append(o)

        # Render all groups
        rendered: list[str | None] = []
        for group_name, options in groups.items():
            if not options:
                continue
            name = f"{group_name or ''} Options".lstrip().capitalize()
            rendered.append(self._format_option_group(name, options))

        return "\n\n".join(group for group in rendered if group)

    def _format_positional_with_mod(self, positional: Config[t.Any]) -> str:
        # E.g.: [FILES]...
        pos_name = positional.name.upper()
        name = f"[{pos_name}]{positional.modifier}"
        return name

    def _format_positional(self, positional: Config[t.Any]) -> tuple[str, ...]:
        # E.g.: [FILES]... or FILES
        name = (
            self.theme.positional(self._format_positional_with_mod(positional))
            if not self.show_option_types
            else self.theme.positional(positional.name.upper())
        )

        help = positional.help or ""
        type_str = (
            self.theme.type_str(str(positional.parser).upper())
            if self.show_option_types
            else ""
        )
        return name, type_str, self._maybe_norm_help(help)

    def _format_positionals(self, positionals: list[Config[t.Any]]) -> str | None:
        if not positionals:
            return None

        name: list[str] = []
        type_str: list[str] = []
        help: list[str] = []
        for p in positionals:
            n, ts, hp = self._format_positional(p)
            name.append(n)
            type_str.append(ts)
            help.append(hp)

        return self._maybe_boxed(name, type_str, help, title="Arguments")

    def _format_subcommand(self, subcmd: type[Command]) -> tuple[str, str]:
        name = self.theme.subcommand(subcmd.prog())
        help = subcmd.help() or ""
        return name, self._maybe_norm_help(help)

    def _format_subcommands(self, subcommands: list[type[Command]]) -> str | None:
        if not subcommands:
            return None

        name: list[str] = []
        help: list[str] = []
        for p in subcommands:
            n, hp = self._format_subcommand(p)
            name.append(n)
            help.append(hp)

        return self._maybe_boxed(name, help, title="Subcommands")

    def _format_header(
        self,
        full_command: list[str],
        options: list[Config[t.Any]],
        positionals: list[Config[t.Any]],
        subcommands: list[type[Command]],
    ) -> str:
        prefix = self.theme.usage("Usage:")
        command_str = self.theme.usage_command(" ".join(full_command))

        positionals_str: list[str] = []
        for pos in positionals:
            name = self._format_positional_with_mod(pos)
            positionals_str.append(self.theme.usage_args(name))
        positional = " " + " ".join(positionals_str) if positionals else ""

        option = self.theme.usage_args(" [OPTIONS]") if options else ""
        command = self.theme.usage_args(" COMMAND") if subcommands else ""

        return f"{prefix} {command_str}{positional}{option}{command}"

    def _format_description(self, description: str | None) -> str | None:
        if not description:
            return None
        return self._maybe_norm_help(description)

    def _format_epilog(self, epilog: str | None) -> str | None:
        if not epilog:
            return None
        return self._maybe_norm_help(epilog)

    def _format_exception(self, exception: Exception | None) -> str | None:
        if not exception:
            return None

        if self.boxed:
            return self._maybe_boxed(
                format_traceback(exception), title="Error", color="red"
            )

        # Special section title since it's an error
        section_title = style("Error:", fg="red", bold=True)
        stacked = "\n".join(indented(format_traceback(exception, color=None)))
        return f"{section_title}\n{stacked}"

    def format_help(
        self,
        full_command: list[str],
        description: str | None,
        epilog: str | None,
        options: list[Config[t.Any]],
        positionals: list[Config[t.Any]],
        subcommands: list[type[Command]],
        exception: Exception | None,
    ) -> str:
        lines: list[str | None] = []

        # Description
        lines.append(self._format_description(description))

        # Header
        lines.append(
            self._format_header(full_command, options, positionals, subcommands)
        )

        # Subcommands
        lines.append(self._format_subcommands(subcommands))

        # Positionals
        lines.append(self._format_positionals(positionals))

        # Options
        lines.append(self._format_options(options))

        # Epilog
        lines.append(self._format_epilog(epilog))

        # Exceptions
        lines.append(self._format_exception(exception))

        joined = "\n\n".join(line for line in lines if line)
        return joined + "\n"



import asyncio
import re
import shutil
import time
from contextlib import suppress
from dataclasses import dataclass
from pathlib import Path
from textwrap import dedent

import anyio
import tomllib

import clypi.parsers as cp
from clypi import Command, Positional, Spinner, arg, boxed, cprint, style

MDTEST_DIR = Path.cwd() / ".mdtest"
PREAMBLE = """\
import clypi
import clypi.parsers as cp
from pathlib import Path
from typing import reveal_type
from clypi import *
from enum import Enum
import asyncio
from datetime import datetime, timedelta
"""


class TestFailed(Exception):
    pass


@dataclass
class Test:
    name: str
    orig: str
    code: str
    args: str
    stdin: str


async def parse_file(sm: asyncio.Semaphore, file: Path) -> list[Test]:
    tests: list[Test] = []
    base_name = file.as_posix().replace("/", "-").replace(".md", "").lower()

    # Wait for turn
    await sm.acquire()

    async with await anyio.open_file(file, "r") as f:
        current_test: list[str] = []
        in_test, args, stdin = False, "", ""
        async for line in f:
            # End of a code block
            if "```" in line and current_test:
                code = "\n".join(current_test[1:])
                tests.append(
                    Test(
                        name=f"{base_name}-{len(tests)}",
                        orig=dedent(code),
                        code=PREAMBLE + dedent(code),
                        args=args,
                        stdin=stdin + "\n",
                    )
                )
                in_test, current_test, args, stdin = False, [], "", ""

            # We're in a test, accumulate all lines
            elif in_test:
                current_test.append(line.removeprefix("> ").removeprefix(">").rstrip())

            # Mdtest arg definition
            elif g := re.search("<!-- mdtest-args (.*) -->", line):
                args = g.group(1)
                in_test = True

            # Mdtest stdin definition
            elif g := re.search("<!-- mdtest-stdin (.*) -->", line):
                stdin = g.group(1)
                in_test = True

            # Mdtest generic definition
            elif g := re.search("<!-- mdtest -->", line):
                in_test = True

    sm.release()
    cprint(style("✔", fg="green") + f" Collected {len(tests)} tests for {file}")
    return tests


def error_msg(test: Test, stdout: str | None = None, stderr: str | None = None) -> str:
    error: list[str] = []
    error.append(style(f"\n\nError running test {test.name!r}\n", fg="red", bold=True))
    error.append(boxed(test.orig, title="Code", width="max"))

    if stdout:
        error.append("")
        error.append(boxed(stdout.strip(), title="Stdout", width="max"))

    if stderr:
        error.append("")
        error.append(boxed(stderr.strip(), title="Stderr", width="max"))

    return "\n".join(error)


class Runner:
    def __init__(self, parallel: int, timeout: int) -> None:
        self.sm = asyncio.Semaphore(parallel)
        self.timeout = timeout

    async def run_test(self, test: Test) -> tuple[str, list[str]]:
        # Save test to temp file
        test_file = MDTEST_DIR / f"{test.name}.py"
        test_file.write_text(test.code)

        commands = [f"uv run --all-extras {test_file}"]
        if test.args:
            commands[0] += f" {test.args}"
        commands.append(f"uv run --all-extras pyright {test_file}")

        # Run the test
        errors: list[str] = []
        for command in commands:
            # Await the subprocess to run it
            proc = await asyncio.create_subprocess_shell(
                command,
                stdout=asyncio.subprocess.PIPE,
                stdin=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
            try:
                stdout, stderr = await proc.communicate(test.stdin.encode())
            except:
                proc.terminate()
                raise

            # If no errors, return
            if proc.returncode == 0:
                continue

            # If there was an error, pretty print it
            errors.append(error_msg(test, stdout.decode(), stderr.decode()))

        if not errors:
            return test.name, []
        return test.name, errors

    async def run_test_with_timeout(self, test: Test) -> tuple[str, list[str]]:
        await self.sm.acquire()
        start = time.perf_counter()
        try:
            async with asyncio.timeout(self.timeout):
                return await self.run_test(test)
        except TimeoutError:
            error = error_msg(
                test,
                stderr=f"Test timed out after {time.perf_counter() - start:.3f}s",
            )
            return test.name, [error]
        finally:
            self.sm.release()

    async def run_mdtests(self, tests: list[Test]) -> int:
        errors: list[str] = []
        async with Spinner("Running Markdown Tests", capture=True) as s:
            coros = [self.run_test_with_timeout(test) for test in tests]
            for task in asyncio.as_completed(coros):
                idx, err = await task
                if not err:
                    cprint(style("✔", fg="green") + f" Finished test {idx}")
                else:
                    errors.extend(err)
                    cprint(style("×", fg="red") + f" Finished test {idx}")

            if errors:
                await s.fail()

        for err in errors:
            cprint(err)

        return 1 if errors else 0


class Mdtest(Command):
    """
    Run python code embedded in markdown files to ensure it's
    runnable.
    """

    files: Positional[list[Path] | None] = arg(
        help="The list of markdown files to test",
        default=None,
    )
    parallel: int | None = arg(None, parser=cp.Int(positive=True))
    timeout: int = arg(4, parser=cp.Int(positive=True))
    config: Path = Path("./pyproject.toml")

    def load_config(self):
        if not self.config.exists():
            return Mdtest()

        with open(self.config, "rb") as f:
            conf = tomllib.load(f)

        with suppress(KeyError):
            data = conf["tool"]["mdtest"]
            parallel = int(data["parallel"]) if "parallel" in data else None
            files = [p for f in data["include"] for p in Path().glob(f)]
            return Mdtest(files, parallel)

        return Mdtest()

    async def run(self) -> None:
        conf = self.load_config()
        files = self.files or conf.files
        parallel = self.parallel or conf.parallel or 1
        if files is None:
            cprint("No files to run!", fg="yellow")
            return

        # Setup test dir
        MDTEST_DIR.mkdir(exist_ok=True)

        # Assert each file exists
        for file in files:
            assert file.exists(), f"File {file} does not exist!"

        try:
            # Collect tests
            async with Spinner("Collecting Markdown Tests", capture=True):
                sm = asyncio.Semaphore(parallel)
                per_file = await asyncio.gather(
                    *(
                        parse_file(sm, file)
                        for file in files
                        if not file.parents[-1].name.startswith(".")
                    )
                )
                all_tests = [test for file in per_file for test in file]

            # Run each file
            print()
            code = await Runner(parallel, self.timeout).run_mdtests(all_tests)
        finally:
            # Cleanup
            shutil.rmtree(MDTEST_DIR)

        raise SystemExit(code)


if __name__ == "__main__":
    mdtest = Mdtest.parse()
    mdtest.start()



from clypi._colors import ColorType, style


class ClypiException(Exception):
    pass


class MaxAttemptsException(ClypiException):
    pass


class AbortException(ClypiException):
    pass


def format_traceback(err: BaseException, color: ColorType | None = "red") -> list[str]:
    # Get the traceback bottom up
    tb: list[BaseException] = [err]
    while tb[-1].__cause__ is not None:
        tb.append(tb[-1].__cause__)

    lines: list[str] = []
    for i, e in enumerate(reversed(tb)):
        icon = "  " * (i - 1) + " ↳ " if i != 0 else ""
        s = style(f"{icon}{str(e)}", fg=color)
        lines.append(s)
    return lines


def print_traceback(err: BaseException) -> None:
    for line in format_traceback(err):
        print(line)



import typing as t

import pytest

from clypi import Command, arg
from tests.cli_parse_test import parametrize


class Run(Command):
    """
    Runs all files
    """

    verbose: bool = arg(inherited=True)
    env: str = arg(inherited=True)


class Main(Command):
    subcommand: Run | None = None
    verbose: bool = arg(False, short="v", help="Whether to show more output")
    env: t.Literal["qa", "prod"] = arg(help="Whether to show more output")


@parametrize(
    "args,expected,fails",
    [
        ([], {}, True),
        (["-v"], {}, True),
        (
            ["-v", "--env", "qa"],
            {
                "verbose": True,
                "env": "qa",
            },
            False,
        ),
        (
            ["--env", "qa", "-v", "run"],
            {
                "verbose": True,
                "env": "qa",
            },
            False,
        ),
        (
            ["--env", "qa", "run", "-v"],
            {
                "verbose": True,
                "env": "qa",
                "run": {"verbose": True, "env": "qa"},
            },
            False,
        ),
        (
            ["run", "--env", "qa", "-v"],
            {
                "verbose": True,
                "env": "qa",
                "run": {"verbose": True, "env": "qa"},
            },
            False,
        ),
        (["run", "-v"], {}, True),
    ],
)
def test_parse_inherited(args: list[str], expected: dict[str, t.Any], fails: bool):
    if fails:
        with pytest.raises(Exception):
            _ = Main.parse(args)
        return

    # Check command
    main = Main.parse(args)
    assert main is not None
    for k, v in expected.items():
        if k == "run":
            continue
        lc_v = getattr(main, k)
        assert lc_v == v, f"{k} should be {v} but got {lc_v}"

    # Check subcommand
    if "run" in expected:
        assert main.subcommand is not None
        assert isinstance(main.subcommand, Run)
        for k, v in expected["run"].items():
            lc_v = getattr(main, k)
            assert lc_v == v, f"run.{k} should be {v} but got {lc_v}"



import typing as t

from clypi._util import UNSET, Unset, visible_width

OverflowStyle = t.Literal["ellipsis", "wrap"]


def wrap(
    s: str, width: int, overflow_style: OverflowStyle | Unset = UNSET
) -> list[str]:
    """
    If a string is larger than width, it either wraps the string into new
    lines or appends an ellipsis
    """
    if visible_width(s) <= width:
        return [s]

    if overflow_style is UNSET:
        from clypi._configuration import get_config

        overflow_style = get_config().overflow_style

    if overflow_style == "ellipsis":
        return [s[: width - 1] + "…"]

    res: list[str] = []
    while s:
        res.append(s[:width])
        s = s[width:].strip()

    return res



import clypi.parsers as cp
from clypi import Command, arg, cprint


class VerboseIntParser(cp.Int):
    """Helper parser to show the user when the parsing is being done"""

    def __call__(self, raw: str | list[str], /) -> int:
        cprint(f"⚠️ The call to to parse {raw} as an int was executed!", fg="yellow")
        return super().__call__(raw)


class Main(Command):
    runner: bool = arg(
        False,
        help="Whether you run",
        prompt="Do you run?",
    )
    often: int = arg(
        help="The frequency you run with in days",
        prompt="How many days a week do you run?",
        defer=True,
        parser=VerboseIntParser(),
    )

    async def run(self):
        print("Command execution started...")

        if not self.runner:
            cprint("You are not a runner!", fg="green", bold=True)
            cprint("Try answering yes on the next try :)", bold=True)
        else:
            cprint(
                # This line will trigger the evaluation of `often` and prompt
                # the user if it was not provided as a CLI arg
                f"You are a runner and run every {self.often} days!",
                fg="green",
                bold=True,
            )
            cprint("Try answering no on the next try :)", bold=True)


if __name__ == "__main__":
    main: Main = Main.parse()
    main.start()



from clypi import stack


def test_one_line_stack():
    result = stack(["a" * 10], ["b" * 10])
    expected = "aaaaaaaaaa  bbbbbbbbbb"
    assert result == expected.strip()


def test_two_line_stack():
    result = stack(["a", "a" * 10], ["b" * 10, "b"], lines=True)
    expected = [
        "a           bbbbbbbbbb",
        "aaaaaaaaaa  b",
    ]
    assert result == expected


def test_overflowing_line_stack():
    result = stack(["a", "a"], ["b" * 12, "b" * 10], max_width=13, lines=True)
    expected = [
        "a  bbbbbbbbbb",
        "   bb",
        "a  bbbbbbbbbb",
    ]
    assert result == expected



from __future__ import annotations

import clypi
import clypi.parsers as cp


def _validate_earth_age(x: str | list[str]) -> int:
    if isinstance(x, str):
        x_int = int(x)
        if x_int == 4_543_000_000:
            return x_int
    raise ValueError("The Earth is 4.543 billion years old. Try 4543000000.")


def main() -> None:
    # Basic prompting
    name = clypi.prompt("What's your name?")

    # Default values
    is_cool = clypi.confirm("Is clypi cool?", default=True)

    # Custom types with parsing
    age = clypi.prompt(
        "How old are you?",
        parser=cp.Int(gte=18),
        hide_input=True,
    )
    hours = clypi.prompt(
        "How many hours are there in a day?",
        parser=cp.Union(cp.TimeDelta(), cp.Int()),
    )

    # Custom validations
    earth = clypi.prompt(
        "How old is The Earth?",
        parser=_validate_earth_age,
    )

    # -----------
    print()
    clypi.cprint("🚀 Summary", bold=True, fg="green")
    answer = clypi.Styler(fg="magenta", bold=True)
    print(" ↳  Name:", answer(name))
    print(" ↳  Clypi is cool:", answer(is_cool))
    print(" ↳  Age:", answer(age))
    print(" ↳  Hours in a day:", answer(hours), f"({type(hours).__name__})")
    print(" ↳  Earth age:", answer(earth))


if __name__ == "__main__":
    main()



from typing import Iterable


def distance(this: str, other: str) -> float:
    """
    Modified version of the Levenshtein distance to consider the case
    of the letters being compared so that dist(a, A) < dist(a, b)
    """
    if not this or not other:
        return max(len(this), len(other))

    n, m = len(this), len(other)
    dist: list[list[float]] = [[0 for _ in range(m + 1)] for _ in range(n + 1)]

    # Prepopulate first X and Y axis
    for t in range(0, n + 1):
        dist[t][0] = t
    for o in range(0, m + 1):
        dist[0][o] = o

    def _subst_dist(t: str, o: str) -> float:
        if t == o:
            return 0
        elif t.lower() == o.lower():
            return 0.5
        return 1

    # Compute actions
    for t in range(n):
        for o in range(m):
            insertion = dist[t][o + 1] + 1
            deletion = dist[t + 1][o] + 1
            substitution = dist[t][o] + _subst_dist(this[t], other[o])
            dist[t + 1][o + 1] = min(insertion, deletion, substitution)

    # Get bottom right of computed matrix
    return dist[n][m]


def closest(word: str, options: Iterable[str]) -> tuple[str, float]:
    """
    Given a word and a list of options, it returns the closest
    option to that word and it's distance
    """
    dists = [distance(word, o) for o in options]
    if not dists:
        return "", float("inf")
    min_opt = min(zip(options, dists), key=lambda x: x[1])
    return min_opt



from pathlib import Path

import pytest
from typing_extensions import override

from clypi import Command, Positional, arg
from clypi._cli.arg_parser import Arg


class ExampleSubCommand(Command):
    """Some sample docs"""

    positional: Positional[tuple[str | Path, ...]]

    async def run(self):
        print("subcommand")


class ExampleCommand(Command):
    """
    Some sample documentation for the main command
    """

    flag: bool = False
    subcommand: ExampleSubCommand | None = None
    option: list[str] = arg(
        short="o", help="A list of strings please", default_factory=list
    )

    @override
    @classmethod
    def prog(cls):
        return "example"

    @override
    @classmethod
    def epilog(cls):
        return "Some text to display after..."

    async def run(self):
        print("main")


def test_expected_base():
    assert ExampleCommand.help() == "Some sample documentation for the main command"
    assert ExampleCommand.prog() == "example"
    assert ExampleCommand.full_command() == ["example"]
    assert ExampleCommand.epilog() == "Some text to display after..."


def test_expected_options():
    opts = ExampleCommand.options()
    assert len(opts) == 2

    assert opts["flag"].name == "flag"
    assert opts["flag"].arg_type is bool
    assert opts["flag"].nargs == 0

    assert opts["option"].name == "option"
    assert opts["option"].arg_type == list[str]
    assert opts["option"].nargs == "*"


def test_expected_positional():
    pos = ExampleSubCommand.positionals()
    assert len(pos) == 1

    assert pos["positional"].name == "positional"
    assert pos["positional"].arg_type == Positional[tuple[str | Path, ...]]
    assert pos["positional"].nargs == "+"


def test_expected_subcommands():
    ec = ExampleCommand.subcommands()
    assert len(ec) == 2

    assert ec[None] is None

    sub = ec["example-sub-command"]
    assert sub is ExampleSubCommand
    assert sub.prog() == "example-sub-command"
    assert sub.help() == "Some sample docs"


def test_expected_cls_introspection():
    assert ExampleCommand.option == []


def test_expected_init():
    cmd = ExampleCommand()
    assert cmd.flag is False
    assert cmd.option == []
    assert cmd.subcommand is None


def test_expected_init_with_kwargs():
    cmd = ExampleCommand(
        flag=True, option=["f"], subcommand=ExampleSubCommand(positional=tuple("g"))
    )
    assert cmd.flag is True
    assert cmd.option == ["f"]
    assert cmd.subcommand is not None
    assert cmd.subcommand.positional == tuple("g")


def test_expected_init_with_args():
    cmd = ExampleCommand(True, ExampleSubCommand(tuple("g")), ["f"])
    assert cmd.flag is True
    assert cmd.option == ["f"]
    assert cmd.subcommand is not None
    assert cmd.subcommand.positional == tuple("g")


def test_expected_init_with_mixed_args_kwargs():
    cmd = ExampleCommand(True, ExampleSubCommand(tuple("g")), option=["f"])
    assert cmd.flag is True
    assert cmd.option == ["f"]
    assert cmd.subcommand is not None
    assert cmd.subcommand.positional == tuple("g")


def test_expected_repr():
    cmd = ExampleCommand(
        flag=True, option=["f"], subcommand=ExampleSubCommand(positional=tuple("g"))
    )
    assert (
        str(cmd)
        == "ExampleCommand(flag=True, option=['f'], subcommand=ExampleSubCommand(positional=('g',)))"
    )


def test_get_similar_opt_error():
    with pytest.raises(ValueError) as exc_info:
        raise ExampleCommand.get_similar_arg_error(
            Arg(
                "falg",  # codespell:ignore
                "--falg",
                "long-opt",
            )
        )

    assert exc_info.value.args[0] == "Unknown option '--falg'. Did you mean '--flag'?"


def test_get_similar_opt_short_error():
    with pytest.raises(ValueError) as exc_info:
        raise ExampleCommand.get_similar_arg_error(
            Arg(
                "c",  # codespell:ignore
                "-c",
                "short-opt",
            )
        )

    assert exc_info.value.args[0] == "Unknown option '-c'. Did you mean '-o'?"


def test_get_similar_subcmd_error():
    with pytest.raises(ValueError) as exc_info:
        raise ExampleCommand.get_similar_arg_error(
            Arg(
                "example-suv-command",
                "example-suv-command",
                "pos",
            )
        )

    assert (
        exc_info.value.args[0]
        == "Unknown argument 'example-suv-command'. Did you mean 'example-sub-command'?"
    )


def test_get_similar_non_similar():
    with pytest.raises(ValueError) as exc_info:
        raise ExampleCommand.get_similar_arg_error(
            Arg(
                "foo",
                "foo",
                "pos",
            )
        )

    assert exc_info.value.args[0] == "Unknown argument 'foo'"



from __future__ import annotations

from typing import Generator

from clypi import ALL_COLORS, ColorType, boxed, stack, style


# --- DEMO UTILS ---
def _all_colors() -> Generator[tuple[ColorType, ...], None, None]:
    mid = len(ALL_COLORS) // 2
    normal, bright = ALL_COLORS[:mid], ALL_COLORS[mid:]
    for color in zip(normal, bright):
        yield color


# --- DEMO START ---
def main() -> None:
    fg_block: list[str] = []
    for color, bright_color in _all_colors():
        fg_block.append(
            style("██ " + color.ljust(9), fg=color)
            + style("██ " + bright_color.ljust(16), fg=bright_color)
        )

    bg_block: list[str] = []
    for color, bright_color in _all_colors():
        bg_block.append(
            style(color.ljust(9), bg=color)
            + " "
            + style(bright_color.ljust(16), bg=bright_color)
        )

    style_block: list[str] = []
    style_block.append(style("I am bold", bold=True))
    style_block.append(style("I am dim", dim=True))
    style_block.append(style("I am underline", underline=True))
    style_block.append(style("I am blink", blink=True))
    style_block.append(style("I am reverse", reverse=True))
    style_block.append(style("I am strikethrough", strikethrough=True))

    stacked_colors = stack(
        boxed(fg_block, title="Foregrounds"),
        boxed(bg_block, title="Backgrounds"),
        boxed(style_block, title="Styles"),
    )
    print(stacked_colors)


if __name__ == "__main__":
    main()



import sys

import clypi
from clypi import ClypiConfig, ClypiFormatter, Command, Styler, Theme, arg, configure
from examples.uv.add import Add
from examples.uv.init import Init
from examples.uv.pip import Pip
from examples.uv.remove import Remove


class Uv(Command):
    """
    A clone of an extremely fast Python package manager.
    """

    subcommand: Add | Init | Pip | Remove | None
    quiet: bool = arg(False, short="q", help="Do not print any output", group="global")
    version: bool = arg(False, short="V", help="Display the uv version", group="global")
    no_cache: bool = arg(
        False,
        help="Avoid reading from or writing to the cache, instead using a temporary directory for the duration of the operation",
        hidden=True,
        group="global",
    )

    async def run(self) -> None:
        # If the version was requested, print it
        if self.version:
            clypi.cprint("clypi's UV 0.0.1", fg="green")
            sys.exit(0)

        if not self.quiet:
            self.print_help()


if __name__ == "__main__":
    # Configure the CLI to look like uv's
    configure(
        ClypiConfig(
            theme=Theme(
                usage=Styler(fg="green", bold=True),
                usage_command=Styler(fg="cyan", bold=True),
                usage_args=Styler(fg="cyan"),
                section_title=Styler(fg="green", bold=True),
                subcommand=Styler(fg="cyan", bold=True),
                long_option=Styler(fg="cyan", bold=True),
                short_option=Styler(fg="cyan", bold=True),
                positional=Styler(fg="cyan"),
                placeholder=Styler(fg="cyan"),
                prompts=Styler(fg="green", bold=True),
            ),
            help_formatter=ClypiFormatter(
                boxed=False,
                show_option_types=False,
            ),
        )
    )

    # Parse and run the commands
    uv = Uv.parse()
    uv.start()



from __future__ import annotations

import asyncio
import dataclasses
import inspect
import logging
import re
import sys
import typing as t
from types import NoneType, UnionType

from clypi import _type_util, parsers
from clypi._cli import arg_config, arg_parser, autocomplete
from clypi._cli.arg_config import Config, Positional, arg
from clypi._cli.context import CurrentCtx
from clypi._cli.deferred import DeferredValue
from clypi._cli.distance import closest
from clypi._cli.formatter import ClypiFormatter, Formatter
from clypi._configuration import get_config
from clypi._exceptions import print_traceback
from clypi._prompts import prompt
from clypi._util import UNSET

logger = logging.getLogger(__name__)

__all__ = (
    "ClypiFormatter",
    "Command",
    "Formatter",
    "Positional",
    "arg",
)


HELP_ARGS: tuple[str, ...] = ("help", "-h", "--help")

CLYPI_OPTIONS = "__clypi_options__"
CLYPI_POSITIONALS = "__clypi_positionals__"
CLYPI_IN_ORDER_FIELD_NAMES = "__clypi_in_order_field_names__"
CLYPI_SUBCOMMANDS = "__clypi_subcommands__"
CLYPI_PARENTS = "__clypi_parents__"


def _camel_to_dashed(s: str):
    return re.sub(r"(?<!^)(?=[A-Z])", "-", s).lower()


class _CommandMeta(type):
    def __init__(
        self,
        name: str,
        bases: tuple[type, ...],
        attrs: dict[str, t.Any],
        /,
        **kwds: t.Any,
    ) -> None:
        super(_CommandMeta, self).__init__(name, bases, attrs)
        self._ensure_fields_are_annotated()
        self._configure_subcommands()
        self._configure_fields()

    @t.final
    def _ensure_fields_are_annotated(self) -> None:
        """
        Ensures that every single field is annotated with type hints
        """
        annotations: dict[str, t.Any] = inspect.get_annotations(self, eval_str=True)
        for name, value in self.__dict__.items():
            if (
                not name.startswith("_")
                and not isinstance(value, classmethod)
                and not callable(value)
                and name not in annotations
            ):
                raise TypeError(f"{name!r} has no type annotation")

    @t.final
    def _configure_fields(self) -> None:
        """
        Parses the type hints from the class extending Command and assigns each
        a field Config with all the necessary info to display and parse them.
        """
        annotations: dict[str, t.Any] = inspect.get_annotations(self, eval_str=True)

        # Mappings for each arg type
        options: dict[str, arg_config.Config[t.Any]] = {}
        positionals: dict[str, arg_config.Config[t.Any]] = {}
        field_names: list[str] = []

        # Get the config for each field
        for field, _type in annotations.items():
            field_names.append(field)
            if field == "subcommand":
                continue

            # Get the specified default for the field (e.g.: foo: bool)
            default = getattr(self, field, UNSET)

            # Check if it comes from `arg()` or if it's a real value
            if isinstance(default, arg_config.PartialConfig):
                field_conf = arg_config.Config.from_partial(
                    partial=default,
                    name=field,
                    parser=default.parser or parsers.from_type(_type),
                    arg_type=_type,
                )
            else:
                # If it's a real value then the default is the value itself
                field_conf = arg_config.Config(
                    name=field,
                    default=default,
                    parser=parsers.from_type(_type),
                    arg_type=_type,
                )

            # Store in the right dict
            if field_conf.is_positional:
                positionals[field] = field_conf
            else:
                options[field] = field_conf

            # Set the values in the class properly instead of keeping the
            # Config fields around
            if not field_conf.has_default() and hasattr(self, field):
                delattr(self, field)
            elif field_conf.has_default():
                setattr(self, field, field_conf.get_default())

        # Store all fields
        setattr(self, CLYPI_OPTIONS, options)
        setattr(self, CLYPI_POSITIONALS, positionals)
        setattr(self, CLYPI_IN_ORDER_FIELD_NAMES, field_names)

    @t.final
    def _configure_subcommands(self) -> None:
        """
        Parses the type hints from the class extending Command and stores the
        subcommand class if any
        """
        annotations: dict[str, t.Any] = inspect.get_annotations(self, eval_str=True)
        if "subcommand" not in annotations:
            return

        _type = annotations["subcommand"]
        subcmds_tmp = [_type]
        if isinstance(_type, UnionType):
            subcmds_tmp = _type_util.union_inner(_type)

        # Store in mapping (name -> type)
        subcmds: dict[str | None, type[Command] | None] = {}
        for v in subcmds_tmp:
            if inspect.isclass(v) and issubclass(v, Command):
                subcmds[v.prog()] = v
            elif v is NoneType:
                subcmds[None] = None
            else:
                raise TypeError(
                    f"Did not expect to see a subcommand {v} of type {type(v)}"
                )

        # Replace the class with a default if it has any
        default = getattr(self, "subcommand", UNSET)
        if default is not UNSET:
            setattr(self, "subcommand", default)
        elif hasattr(self, "subcommand"):
            delattr(self, "subcommand")

        # Store list of subcommands
        setattr(self, CLYPI_SUBCOMMANDS, subcmds)

    @t.final
    def inherit(self, parent: type[Command]) -> list[str]:
        """
        This function is called by the parent command during parsing to configure
        inherited fields and the parenthood relationship so that the full command
        is displayed properly.

        Returns the list of inherited fields
        """
        setattr(self, CLYPI_PARENTS, parent.full_command())

        # For inherited args, configure them with the parent's configs
        options = self.options()
        inherited: list[str] = []
        for opt, opt_config in parent.options().items():
            if opt not in options or not options[opt].inherited:
                continue
            options[opt] = dataclasses.replace(
                opt_config,
                # Keep inherited and group config
                inherited=True,
                group=options[opt].group or opt_config.group,
            )
            inherited.append(opt)

        return inherited

    @t.final
    def subcommands(self) -> dict[str | None, type[Command] | None]:
        return getattr(self, CLYPI_SUBCOMMANDS, None) or {None: None}

    @t.final
    def options(self) -> dict[str, Config[t.Any]]:
        return getattr(self, CLYPI_OPTIONS, {})

    @t.final
    def positionals(self) -> dict[str, Config[t.Any]]:
        return getattr(self, CLYPI_POSITIONALS, {})

    @t.final
    def field_names(self) -> list[str]:
        return getattr(self, CLYPI_IN_ORDER_FIELD_NAMES, [])

    @t.final
    def get_field_conf(self, name: str) -> Config[t.Any]:
        """
        Helper function to find a field config with a given name. We need this
        since we store the args in different dicts yet we want to iterate through
        them in order to parse them how the user specified.
        """
        if conf := self.positionals().get(name):
            return conf
        if conf := self.options().get(name):
            return conf
        raise ValueError(f"Unknown field {name}")


@t.dataclass_transform()
class Command(metaclass=_CommandMeta):
    def __init__(self, *args: t.Any, **kwargs: t.Any) -> None:
        """
        Implementation of the init function for Commands. Similar to
        dataclasses, we only check for the field's existence and for the
        order the fields are provided in, but we do not check the data type
        for the actual provided values due to complication of checking types
        from type hints to real variables.
        """
        fields: dict[str, t.Any] = {}

        # Reversed so that we can `.pop()` in order
        field_names = list(reversed(self.__class__.field_names()))
        arg_ls = list(reversed(args))

        # From *args
        while arg_ls:
            fields[field_names.pop()] = arg_ls.pop()

        # From *kwargs
        for k, v in kwargs.items():
            if k in fields:
                raise TypeError(
                    f"Found duplicate field {k} for {self.__class__.__name__}"
                )
            if k not in field_names:
                raise TypeError(f"Invalid argument {k} for {self.__class__.__name__}")

            fields[k] = v
            field_names.remove(k)

        # Validate all fields and populate defaults
        validated = self._validate_fields(fields, name=self.__class__.__name__)

        # Save all fields to current instance
        for k, v in validated.items():
            setattr(self, k, v)

    @classmethod
    def _validate_fields(cls, fields: dict[str, t.Any], name: str) -> dict[str, t.Any]:
        """
        Takes in a dictionary of field names and values, checks if the fields are all
        part of the current Command class, and populates default fields. Raises if any
        field is missing or not a valid one.
        """

        field_map: dict[str, t.Any] = {}

        # Reversed so that we can `.pop()` in order
        field_names = set(cls.field_names())

        # From *kwargs
        for field, value in fields.items():
            if field not in field_names:
                raise TypeError(f"Invalid argument {field} for {name!r}")

            field_map[field] = value
            field_names.remove(field)

        # Check if subcommand is missing
        if "subcommand" in field_names:
            # If command is optional, set it to None
            if None in cls.subcommands():
                field_names.remove("subcommand")
                field_map["subcommand"] = None
            else:
                raise TypeError(f"Missing required subcommand for {name!r}")

        # Set defaults for any other missing fields
        missing_field_names: list[str] = []
        for field in field_names:
            conf = cls.get_field_conf(field)
            if conf.has_default():
                field_map[field] = conf.get_default()
            else:
                missing_field_names.append(field)

        # The user did not provide all of the necessary fields
        if missing_field_names:
            raise TypeError(
                f"Missing required arguments {', '.join(missing_field_names)} for {name!r}"
            )

        return field_map

    @classmethod
    def prog(cls) -> str:
        """
        The name of the command being executed. E.g.: install
        """
        return _camel_to_dashed(cls.__name__)

    @t.final
    @classmethod
    def full_command(cls) -> list[str]:
        """
        The full path to the current command being ran. E.g.: pip install
        """
        return cls.parents() + [cls.prog()]

    @classmethod
    def epilog(cls) -> str | None:
        """
        Optionally define text to display after the help message
        """
        return None

    @classmethod
    def parents(cls) -> list[str]:
        """
        A list of parent commands for this command. E.g.: pip
        """
        return getattr(cls, CLYPI_PARENTS, [])

    @t.final
    @classmethod
    def help(cls):
        """
        A brief description for the command
        """
        doc = inspect.getdoc(cls) or ""
        return doc.replace("\n", " ")

    async def pre_run_hook(self) -> None:
        """
        This function will execute on every command and subcommand right
        before running it.
        """
        pass

    async def post_run_hook(self, exception: Exception | None) -> None:
        """
        This function will execute on every command and subcommand right
        after running it.
        """
        pass

    async def run(self) -> None:
        """
        This function is where the business logic of your command
        should live.

        `self` contains the arguments for this command you can access
        as you would do with any other instance property.
        """
        self.print_help()

    @t.final
    async def astart(self) -> Exception | None:
        exc: Exception | None = None

        # If the class has a subcommand, traverse into the child cmd
        if subcommand := getattr(self, "subcommand", None):
            await self.pre_run_hook()
            try:
                exc = await subcommand.astart()
            except Exception as e:
                exc = e
                raise
            finally:
                await self.post_run_hook(exc)

        # Otherwise this is the command to run, run it
        else:
            await self.pre_run_hook()
            try:
                await self.run()
            except get_config().nice_errors as e:
                exc = e
                print_traceback(e)
            except Exception as e:
                exc = e
                raise
            finally:
                await self.post_run_hook(exc)

        return exc

    @t.final
    def start(self) -> Exception | None:
        return asyncio.run(self.astart())

    @t.final
    @classmethod
    def _next_positional(cls, kwargs: dict[str, t.Any]) -> Config[t.Any] | None:
        """
        Traverse the current collected arguments and find the next positional
        arg we can assign to.
        """
        for name, pos in cls.positionals().items():
            # List positionals are a catch-all
            if _type_util.is_list(pos.arg_type):
                return pos

            if name not in kwargs:
                return pos

        return None

    @t.final
    @classmethod
    def _get_long_name(cls, short: str) -> str | None:
        for field, field_conf in cls.options().items():
            if field_conf.short == short:
                return field
        return None

    @t.final
    @classmethod
    def get_similar_arg_error(cls, arg: arg_parser.Arg) -> ValueError:
        """
        Utility function to find arguments similar to the one the
        user passed in to correct typos.
        """
        similar = None

        if arg.is_pos():
            all_pos: list[str] = [
                *[s for s in cls.subcommands() if s],
                *list(cls.positionals()),
            ]
            pos, dist = closest(arg.value, all_pos)
            # 2 is ~good for typos (e.g.: this -> that)
            if dist <= 2:
                similar = pos
        else:
            all_pos: list[str] = [
                *list(cls.options()),
                *[o.short for o in cls.options().values() if o.short],
            ]
            pos, dist = closest(arg.value, all_pos)
            # 2 is ~good for typos (e.g.: this -> that)
            if dist <= 2:
                similar = f"--{pos}" if len(pos) > 1 else f"-{pos}"

        what = "argument" if arg.is_pos() else "option"
        error = f"Unknown {what} {arg.orig!r}"
        if similar is not None:
            error += f". Did you mean {similar!r}?"

        return ValueError(error)

    @t.final
    @classmethod
    def _safe_parse(
        cls,
        args: t.Iterator[str],
        parent_attrs: dict[str, str | list[str]] | None = None,
    ) -> t.Self:
        """
        Tries parsing args and if an error is shown, it displays the subcommand
        that failed the parsing's help page.
        """
        try:
            return cls._parse(args, parent_attrs)
        except parsers.CATCH_ERRORS as e:
            if not get_config().help_on_fail:
                raise

            # The user might have started typing a subcommand but not
            # finished it so we cannot fully parse it, but we can recommend
            # the current command's args to autocomplete it
            if autocomplete.get_autocomplete_args() is not None:
                autocomplete.list_arguments(cls)

            # Otherwise, help page
            cls.print_help(exception=e)

    @t.final
    @classmethod
    def _parse(
        cls,
        args: t.Iterator[str],
        parent_attrs: dict[str, str | list[str]] | None = None,
    ) -> t.Self:
        """
        Given an iterator of arguments we recursively parse all options, arguments,
        and subcommands until the iterator is complete.

        When we encounter a subcommand, we parse all the types, then try to keep parsing the
        subcommand whilst we assign all inherited types.
        """
        parent_attrs = parent_attrs or {}

        # An accumulator to store unparsed arguments for this class
        unparsed: dict[str, str | list[str]] = {}

        # The current option or positional arg being parsed
        current_attr: CurrentCtx = CurrentCtx()

        def flush_ctx():
            nonlocal current_attr
            if current_attr and current_attr.needs_more():
                raise ValueError(f"Not enough values for {current_attr.name}")
            elif current_attr:
                unparsed[current_attr.name] = current_attr.collected
                current_attr = CurrentCtx()

        # The subcommand we need to parse
        subcommand: type[Command] | None = None

        # If the user is trying to display the help page we can skip some parts
        requested_help = sys.argv[-1].lower() in HELP_ARGS

        # Parse the cmd line arguments
        for a in args:
            parsed = arg_parser.parse_as_attr(a)

            # If we've reached -h or --help
            if parsed.orig.lower() in HELP_ARGS:
                cls.print_help()

            # Try to parse as a subcommand
            if parsed.is_pos() and parsed.value in cls.subcommands():
                subcommand = cls.subcommands()[parsed.value]
                break

            # ---- Try to set to the current option ----
            is_valid_long = parsed.is_long_opt() and parsed.value in cls.options()
            is_valid_short = parsed.is_short_opt() and cls._get_long_name(parsed.value)
            if parsed.is_opt() and not (is_valid_long or is_valid_short):
                raise cls.get_similar_arg_error(parsed)

            if is_valid_long or is_valid_short:
                long_name = cls._get_long_name(parsed.value) or parsed.value
                option = cls.options()[long_name]
                flush_ctx()

                # Boolean flags don't need to parse more args later on
                if option.nargs == 0:
                    unparsed[long_name] = "yes"
                else:
                    current_attr = CurrentCtx(option.name, option.nargs, option.nargs)
                continue

            # Try to assign to the current positional
            if not current_attr.name and (pos := cls._next_positional(unparsed)):
                current_attr = CurrentCtx(pos.name, pos.nargs, pos.nargs)

            # Try to assign to the current ctx
            if current_attr.name and current_attr.has_more():
                current_attr.collect(parsed.value)
                if not current_attr.has_more():
                    flush_ctx()
                continue

            raise cls.get_similar_arg_error(parsed)

        # Flush the context after the loop in case anything is left uncollected
        flush_ctx()

        # If the user requested help, skip prompting/parsing
        parsed_kwargs: dict[str, t.Any] = {}
        if not requested_help:
            # --- Parse as the correct values ---
            for field in cls.field_names():
                if field == "subcommand":
                    continue
                field_conf = cls.get_field_conf(field)

                # If the field was provided through args
                if field in unparsed:
                    parsed_kwargs[field] = field_conf.parser(unparsed[field])

                # If the field comes from a parent command, use that
                elif field_conf.inherited and field in parent_attrs:
                    parsed_kwargs[field] = parent_attrs[field]

                # If the field was not provided but we can defer prompting for the
                # value, we instead set it to a DeferrableValue
                elif field_conf.prompt is not None and field_conf.defer:
                    parsed_kwargs[field] = DeferredValue(
                        prompt=field_conf.prompt,
                        parser=field_conf.parser,
                        default=field_conf.default,
                        default_factory=field_conf.default_factory,
                        hide_input=field_conf.hide_input,
                        max_attempts=field_conf.max_attempts,
                    )

                # If the field was not provided but we can prompt, do so
                elif field_conf.prompt is not None:
                    parsed_kwargs[field] = prompt(
                        field_conf.prompt,
                        default=field_conf.default,
                        default_factory=field_conf.default_factory,
                        hide_input=field_conf.hide_input,
                        max_attempts=field_conf.max_attempts,
                        parser=field_conf.parser,
                    )

                # Otherwise, if the field has a default, use that
                elif field_conf.has_default():
                    parsed_kwargs[field] = field_conf.get_default()

        # Parse the subcommand passing in the parsed types
        if subcommand:
            # Configure parenthood and inherited args
            inherited_fields = subcommand.inherit(cls)

            # Parse the subcommand
            subcmd_instance = subcommand._safe_parse(args, parent_attrs=parsed_kwargs)
            parsed_kwargs["subcommand"] = subcmd_instance

            # If any fields were inherited by the subcommand and populated there,
            # get the same value for this instance
            for inh_field in inherited_fields:
                parsed_kwargs[inh_field] = getattr(subcmd_instance, inh_field)

        # Initialize the instance
        validated = cls._validate_fields(parsed_kwargs, name=cls.prog())
        return cls(**validated)

    @t.final
    @classmethod
    def parse(cls, args: t.Sequence[str] | None = None) -> t.Self:
        """
        Entry point of the program. Depending on some env vars it
        will either run the user-defined program or instead output the necessary
        completions for shells to provide autocomplete
        """
        args = args if args is not None else sys.argv[1:]
        if autocomplete.requested_autocomplete_install(args):
            autocomplete.install_autocomplete(cls)

        # If this is an autocomplete call, we need the args from the env var passed in
        # by the shell's complete function
        if auto_args := autocomplete.get_autocomplete_args():
            args = auto_args

        norm_args = arg_parser.normalize_args(args)
        args_iter = iter(norm_args)
        instance = cls._safe_parse(args_iter)
        if autocomplete.get_autocomplete_args() is not None:
            autocomplete.list_arguments(cls)
        if list(args_iter):
            raise ValueError(f"Unknown arguments {list(args_iter)}")

        return instance

    @classmethod
    def print_help(cls, exception: Exception | None = None) -> t.NoReturn:
        help_str = get_config().help_formatter.format_help(
            full_command=cls.full_command(),
            description=cls.help(),
            epilog=cls.epilog(),
            options=list(cls.options().values()),
            positionals=list(cls.positionals().values()),
            subcommands=[s for s in cls.subcommands().values() if s],
            exception=exception,
        )
        sys.stdout.write(help_str)
        sys.stdout.flush()
        sys.exit(1 if exception else 0)

    def __repr__(self) -> str:
        fields = ", ".join(
            f"{k}={v}"
            for k, v in vars(self).items()
            if v is not None and not k.startswith("_")
        )
        return f"{self.__class__.__name__}({fields})"



from pathlib import Path

import clypi
from clypi import Command, Positional, arg


class Init(Command):
    """Create a new project"""

    path: Positional[Path] = arg(help="The path to use for the project/script")
    name: str = arg(
        help="The name of the project",
        prompt="What's the name of your project/script?",
    )
    description: str = arg(
        help="Set the project description",
        prompt="What's your project/script's description?",
    )

    async def run(self) -> None:
        clypi.cprint("Running `uv init` command...", fg="blue")



from clypi import Command, arg
from clypi.parsers import Str
from tests.prompt_test import replace_stdin


class DeferredError(Exception):
    pass


class Signal:
    called: bool = False

    def __bool__(self):
        return self.called


class SignalParser(Str):
    """Utility parser that just raises to test when it is evaluated"""

    def __init__(self, signal: Signal) -> None:
        self._signal = signal

    def __call__(self, raw: str | list[str], /) -> str:
        self._signal.called = True
        return super().__call__(raw)


def test_never_evaluates():
    called = Signal()

    class Main(Command):
        verbose: bool = False
        some_arg: str = arg(
            prompt="What's the value?",
            defer=True,
            parser=SignalParser(called),
        )

    cmd = Main.parse([])
    assert cmd.verbose is False
    assert not called


def test_defer_with_cli_args():
    called = Signal()

    class Main(Command):
        verbose: bool = False
        some_arg: str = arg(
            prompt="What's the value?",
            defer=True,
            parser=SignalParser(called),
        )

    # Happens during CLI arg parse
    assert not called
    Main.parse(["--some-arg", "foo"])
    assert called


def test_defer_not_provided():
    called = Signal()

    class Main(Command):
        verbose: bool = False
        some_arg: str = arg(
            prompt="What's the value?",
            defer=True,
            parser=SignalParser(called),
        )

    cmd = Main.parse([])

    # Happens during attribute access
    assert not called
    with replace_stdin("foo"):
        assert cmd.some_arg == "foo"
    assert called



import clypi
from clypi import Command, Positional, arg


class Remove(Command):
    """Remove dependencies from the project"""

    packages: Positional[list[str]] = arg(
        help="The names of the dependencies to remove (e.g., `ruff`)"
    )
    dev: bool = arg(
        False, help="Remove the packages from the development dependency group"
    )

    # Inherited opts
    quiet: bool = arg(inherited=True)
    version: bool = arg(inherited=True)
    no_cache: bool = arg(inherited=True)

    async def run(self) -> None:
        clypi.cprint("Running `uv remove` command...", fg="blue")

        # Remove the packages passed as args
        clypi.cprint("\nRemoved packages", fg="blue", bold=True)
        for p in self.packages:
            icon = clypi.style("-", fg="red", bold=True)
            print(f"[{icon}] {p} 0.1.0")



import os
import shlex
import sys
import typing as t

from typing_extensions import override

from clypi import ClypiFormatter, Command, Positional, arg, cprint, get_config, style


class Run(Command):
    """
    Runs all files
    """

    files: Positional[list[str]] = arg(help="The files to run")
    verbose: bool = arg(inherited=True, group="global")
    env: str = arg(inherited=True, group="global")

    async def run(self):
        cprint("Running with:", fg="blue", bold=True)
        cprint(f" - Files: {self.files}", fg="blue")
        cprint(f" - Verbose: {self.verbose}", fg="blue")
        cprint(f" - Env: {self.env}", fg="blue")
        cprint("Done!", fg="green", bold=True)


class Main(Command):
    """
    4ward is an example of how we can reuse args across commands using Clypi.
    """

    subcommand: Run | None = None
    verbose: bool = arg(False, short="v", help="Whether to show more output")
    env: t.Literal["qa", "prod"] = arg(help="Whether to show more output")

    @override
    @classmethod
    def prog(cls):
        return "4ward"

    @override
    @classmethod
    def epilog(cls):
        return "Learn more at http://4ward.org"

    async def run(self):
        cprint("Running with:", fg="blue", bold=True)
        cprint(f" - Verbose: {self.verbose}", fg="blue")
        cprint(f" - Env: {self.env}", fg="blue")
        cprint("Done!", fg="green", bold=True)


if __name__ == "__main__":
    show_inherited = True
    if os.getenv("SHOW_INHERITED") != "1":
        cprint(
            "Not showing inherited args. Try using: "
            + style(
                "SHOW_INHERITED=1 uv run -m examples.cli_inherited "
                + shlex.join(sys.argv[1:]),
                bold=True,
            )
            + "\n\n",
            fg="yellow",
        )
        show_inherited = False

    get_config().help_formatter = ClypiFormatter(show_inherited_options=show_inherited)

    main: Main = Main.parse()
    main.start()



from clypi import parsers
from clypi._cli.arg_config import Positional, arg
from clypi._cli.distance import closest, distance
from clypi._cli.formatter import ClypiFormatter, Formatter
from clypi._cli.main import Command
from clypi._colors import ALL_COLORS, ColorType, Styler, cprint, style
from clypi._components.align import AlignType, align
from clypi._components.boxed import Boxes, boxed
from clypi._components.indented import indented
from clypi._components.separator import separator
from clypi._components.spinners import Spin, Spinner, spinner
from clypi._components.stack import stack
from clypi._components.wraps import OverflowStyle, wrap
from clypi._configuration import ClypiConfig, Theme, configure, get_config
from clypi._exceptions import (
    AbortException,
    ClypiException,
    MaxAttemptsException,
    format_traceback,
    print_traceback,
)
from clypi._prompts import (
    confirm,
    prompt,
)
from clypi.parsers import Parser

__all__ = (
    "ALL_COLORS",
    "AbortException",
    "AlignType",
    "Boxes",
    "ClypiConfig",
    "ClypiException",
    "ClypiFormatter",
    "ColorType",
    "Command",
    "Formatter",
    "MaxAttemptsException",
    "OverflowStyle",
    "Parser",
    "Positional",
    "Spin",
    "Spinner",
    "Styler",
    "Theme",
    "align",
    "arg",
    "boxed",
    "closest",
    "configure",
    "confirm",
    "cprint",
    "distance",
    "format_traceback",
    "get_config",
    "indented",
    "parsers",
    "print_traceback",
    "prompt",
    "separator",
    "spinner",
    "stack",
    "style",
    "wrap",
)



from __future__ import annotations

import datetime as dt
import io
import re
import sys
import typing as t
from contextlib import contextmanager, suppress
from datetime import date, datetime, timedelta
from enum import Enum

import pytest
from pytest import mark

import clypi
from clypi import AbortException, MaxAttemptsException, get_config
from clypi._colors import remove_style


@contextmanager
def replace_stdin(answers: list[str] | str = []):
    answers = [answers] if isinstance(answers, str) else answers
    target = "\n".join(answers)

    orig = sys.stdin
    sys.stdin = io.StringIO(target)
    yield
    sys.stdin = orig


@contextmanager
def replace_stdout():
    orig = sys.stdout
    new = io.StringIO()
    sys.stdout = new
    yield new
    sys.stdout = orig


def _escape_ansi(line: str) -> str:
    """
    https://stackoverflow.com/questions/14693701/how-can-i-remove-the-ansi-escape-sequences-from-a-string-in-python
    """
    ansi_escape = re.compile(r"(?:\x1B[@-_]|[\x80-\x9F])[0-?]*[ -/]*[@-~]")
    return ansi_escape.sub("", line)


def assert_prompted_times(prompted: io.StringIO, times: int):
    __tracebackhide__ = True
    text = _escape_ansi(prompted.getvalue())
    lines = list(filter(None, text.split(": ")))
    assert len(lines) == times


def _raise_error(x: str | list[str]) -> None:
    raise ValueError(f"Invalid number {x}")


class _TestEnum(Enum):
    QA = 1
    PROD = 2


class TestCase:
    def setup_method(self):
        conf = get_config()
        conf.help_on_fail = False
        conf.nice_errors = tuple()

    @mark.parametrize(
        "expected",
        [True, False],
    )
    def test_confirm_with_default(self, expected: bool):
        with replace_stdin("\n") as _:
            assert clypi.confirm("What's your name?", default=expected) == expected

    @mark.parametrize(
        "answer, expected",
        [
            ("y", True),
            ("Yes", True),
            ("true", True),
            ("N", False),
            ("no", False),
            ("False", False),
        ],
    )
    def test_confirm(self, answer: str, expected: bool):
        with replace_stdin(answer) as _:
            assert clypi.confirm("What's your name?") == expected

    def test_confirm_aborts(self):
        with replace_stdin("n") as _:
            with pytest.raises(AbortException):
                _ = clypi.confirm("What's your name?", abort=True)

    @mark.parametrize(
        "answer,expected",
        [
            ("Alice\n", "Alice"),
            ("\n", "John Doe"),
        ],
    )
    def test_prompt_with_default(self, answer: str, expected: str):
        with replace_stdin(answer) as _:
            assert clypi.prompt("What's your name?", default="John Doe") == expected

    @mark.parametrize(
        "answers,expected,times",
        [
            (["Alice"], "Alice", 1),
            (["", "", "Alice"], "Alice", 3),
        ],
    )
    def test_prompt_with_no_default(
        self, answers: list[str], expected: str, times: int
    ):
        with replace_stdin(answers) as _, replace_stdout() as stdout:
            assert clypi.prompt("What's your name?") == expected
            assert_prompted_times(stdout, times)

    def test_prompt_with_parser(self):
        with replace_stdin("42") as _:
            res = clypi.prompt("Some prompt", parser=int)
            assert isinstance(res, int)

    @mark.parametrize(
        "answer,parser",
        [
            ("Alice", int),
            ("42 days", int),
            ("42 asd", timedelta),
            ("42", timedelta),
            ("202-01-01", date),
            ("2021-01-01T00:00", datetime),
            ("1 q", timedelta | None),
        ],
        ids=[
            "Str as Int",
            "TimeDelta as Int",
            "Invalid TimeDelta",
            "Int as TimeDelta",
            "Invalid Date",
            "Invalid DateTime",
            "Invalid TimeDelta or None",
        ],
    )
    def test_prompt_with_parser_fails(self, answer: str, parser: type):
        with replace_stdin(answer) as _, pytest.raises(MaxAttemptsException):
            clypi.prompt("Some prompt", parser=parser, max_attempts=1)

    def test_prompt_with_good_parser(self):
        with replace_stdin("2") as _:
            res = clypi.prompt(
                "Some prompt",
                parser=lambda x: int(x) * 2 if isinstance(x, str) else len(x),
            )
            assert res == 4

    def test_prompt_with_bad_validate(self):
        with replace_stdin("2") as _, pytest.raises(MaxAttemptsException):
            clypi.prompt("Some prompt", parser=_raise_error, max_attempts=1)

    @mark.parametrize(
        "default,expected",
        [
            (False, "y/N"),
            (True, "Y/n"),
            (_TestEnum.QA, "qa"),
            (_TestEnum.PROD, "prod"),
            (dt.datetime(2025, 1, 1), "2025-01-01 00:00:00"),
            (dt.timedelta(days=1), "1 day, 0:00:00"),
        ],
    )
    def test_prompt_default_repr(self, default: t.Any, expected: str):
        with (
            replace_stdin(["", ""]) as _,
            replace_stdout() as stdout,
            suppress(MaxAttemptsException),
        ):
            clypi.prompt("Some prompt", max_attempts=1, default=default)
            assert remove_style(stdout.getvalue()) == f"Some prompt [{expected}]: "



import re
import typing as t
from dataclasses import dataclass


def dash_to_snake(s: str) -> str:
    return re.sub(r"^-+", "", s).replace("-", "_")


def snake_to_dash(s: str) -> str:
    return s.replace("_", "-")


def normalize_args(args: t.Sequence[str]) -> list[str]:
    new_args: list[str] = []
    for a in args:
        if a.startswith("-") and "=" in a:
            new_args.extend(a.split("=", 1))
        else:
            new_args.append(a)
    return new_args


@dataclass
class Arg:
    value: str
    orig: str
    arg_type: t.Literal["long-opt", "short-opt", "pos"]

    def is_pos(self):
        return self.arg_type == "pos"

    def is_long_opt(self):
        return self.arg_type == "long-opt"

    def is_short_opt(self):
        return self.arg_type == "short-opt"

    def is_opt(self):
        return self.is_long_opt() or self.is_short_opt()


def parse_as_attr(arg: str) -> Arg:
    if arg.startswith("--"):
        return Arg(value=dash_to_snake(arg), orig=arg, arg_type="long-opt")

    if arg.startswith("-"):
        return Arg(value=dash_to_snake(arg), orig=arg, arg_type="short-opt")

    return Arg(value=arg, orig=arg, arg_type="pos")



import typing as t
from pathlib import Path

import clypi
import clypi.parsers as cp

t.assert_type(clypi.confirm("A"), bool)
t.assert_type(clypi.prompt("A"), str)
t.assert_type(clypi.prompt("A", parser=int), int)
t.assert_type(clypi.prompt("A", parser=cp.Path()), Path)
t.assert_type(
    clypi.prompt("A", parser=cp.Union(cp.Path(), cp.Int())),
    Path | int,
)



from __future__ import annotations

import enum
import re
import typing as t
from abc import ABC, abstractmethod
from contextlib import suppress
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from pathlib import Path as _Path

from clypi import _type_util as tu
from clypi._exceptions import ClypiException

T = t.TypeVar("T", covariant=True)
X = t.TypeVar("X")
Y = t.TypeVar("Y")


class UnparseableException(ClypiException):
    pass


class Parser(t.Protocol[T]):
    def __call__(self, raw: str | list[str], /) -> T: ...


CATCH_ERRORS: tuple[type[Exception], ...] = (
    ValueError,
    TypeError,
    IndexError,
    ClypiException,
)


class CannotParseAs(ClypiException):
    def __init__(self, value: t.Any, parser: Parser[t.Any]) -> None:
        message = f"Cannot parse {value!r} as {parser}"
        super().__init__(message)


class ClypiParser(ABC, t.Generic[X]):
    name: str | None = None

    @abstractmethod
    def __call__(self, raw: str | list[str], /) -> X: ...

    def __or__(self, other: ClypiParser[Y]) -> Union[X, Y]:
        return Union(self, other)

    def __eq__(self, other: t.Any):
        if not isinstance(other, ClypiParser):
            return False
        if self.__class__ != other.__class__:
            return False
        # TODO: fix this comparison. Only for tests so it's ok for now
        return str(self) == str(other)


def a(cond: bool, value: t.Any, error_msg: str):
    """
    Short hand for assertions with error messages
    """
    if not cond:
        raise ValueError(f"{value} {error_msg}")


@dataclass
class Int(ClypiParser[int]):
    gt: int | None = None
    gte: int | None = None
    lt: int | None = None
    lte: int | None = None
    max: int | None = None
    min: int | None = None
    positive: bool = False
    nonpositive: bool = False
    negative: bool = False
    nonnegative: bool = False

    def __call__(self, raw: str | list[str], /) -> int:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        if int(raw) != float(raw):
            raise ValueError(f"The value {raw!r} is not a valid integer.")

        parsed = int(raw)
        if self.gt is not None:
            a(parsed > self.gt, parsed, f"is not greater than {self.gt}")
        if self.gte is not None:
            a(parsed >= self.gte, parsed, f"is not greater than or equal to {self.gte}")
        if self.lt is not None:
            a(parsed < self.lt, parsed, f"is not less than {self.lt}")
        if self.lte is not None:
            a(parsed <= self.lte, parsed, f"is not less than or equal to {self.lte}")
        if self.max is not None:
            a(parsed <= self.max, parsed, f"is not less than or equal to {self.max}")
        if self.min is not None:
            a(parsed >= self.min, parsed, f"is not greater than or equal to {self.min}")
        if self.positive:
            a(parsed > 0, parsed, "is not positive")
        if self.nonpositive:
            a(parsed <= 0, parsed, "is not non-positive")
        if self.negative:
            a(parsed < 0, parsed, "is not negative")
        if self.nonnegative:
            a(parsed >= 0, parsed, "is non non-negative")

        return parsed

    def __repr__(self) -> str:
        return "integer"


@dataclass
class Float(ClypiParser[float]):
    gt: float | None = None
    gte: float | None = None
    lt: float | None = None
    lte: float | None = None
    max: float | None = None
    min: float | None = None
    positive: bool = False
    nonpositive: bool = False
    negative: bool = False
    nonnegative: bool = False

    def __call__(self, raw: str | list[str], /) -> float:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)

        parsed = float(raw)
        if self.gt is not None:
            a(parsed > self.gt, parsed, f"is not greater than {self.gt}")
        if self.gte is not None:
            a(parsed >= self.gte, parsed, f"is not greater than or equal to {self.gte}")
        if self.lt is not None:
            a(parsed < self.lt, parsed, f"is not less than {self.lt}")
        if self.lte is not None:
            a(parsed <= self.lte, parsed, f"is not less than or equal to {self.lte}")
        if self.max is not None:
            a(parsed <= self.max, parsed, f"is not less than or equal to {self.max}")
        if self.min is not None:
            a(parsed >= self.min, parsed, f"is not greater than or equal to {self.min}")
        if self.positive:
            a(parsed > 0, parsed, "is not positive")
        if self.nonpositive:
            a(parsed <= 0, parsed, "is not non-positive")
        if self.negative:
            a(parsed < 0, parsed, "is not negative")
        if self.nonnegative:
            a(parsed >= 0, parsed, "is non non-negative")
        return parsed

    def __repr__(self) -> str:
        return "float"


class Bool(ClypiParser[bool]):
    TRUE_BOOL_STR_LITERALS: set[str] = {"true", "yes", "y"}
    FALSE_BOOL_STR_LITERALS: set[str] = {"false", "no", "n"}

    def __call__(self, raw: str | list[str], /) -> bool:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)

        raw_lower = raw.lower()
        both = self.TRUE_BOOL_STR_LITERALS | self.FALSE_BOOL_STR_LITERALS
        if raw_lower not in both:
            raise ValueError(
                f"The string {raw!r} is not valid boolean! The only allowed values are: {both}."
            )
        return raw_lower in self.TRUE_BOOL_STR_LITERALS

    def __repr__(self):
        return "{yes|no}"

    def _parts(self):
        """
        Required so that it can be flattened when inside unions or literals
        """
        return ["yes", "no"]


@dataclass
class Str(ClypiParser[str]):
    length: int | None = None
    max: int | None = None
    min: int | None = None
    startswith: str | None = None
    endswith: str | None = None
    regex: str | None = None
    regex_group: int | None = None

    def __call__(self, raw: str | list[str], /) -> str:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)

        if self.length is not None:
            a(len(raw) == self.length, len, f"'s length is not {self.length}")
        if self.max is not None:
            a(len(raw) <= self.max, len, f"'s length is not less than {self.max}")
        if self.min is not None:
            a(len(raw) >= self.min, len, f"'s length is not greater than {self.min}")
        if self.startswith is not None:
            a(
                raw.startswith(self.startswith),
                raw,
                f"does not start with {self.startswith}",
            )
        if self.endswith is not None:
            a(raw.endswith(self.endswith), raw, f"does not end with {self.endswith}")
        if self.regex is not None:
            m = re.search(self.regex, raw)
            a(m is not None, raw, f"does not match the regular expression {self.regex}")
            if m and self.regex_group is not None:
                val = m.group(self.regex_group)
                assert isinstance(val, str)
                raw = val

        return raw

    def __repr__(self) -> str:
        return "text"


@dataclass
class DateTime(ClypiParser[datetime]):
    tz: timezone | None = None

    def __call__(self, raw: str | list[str], /) -> datetime:
        from dateutil.parser import parse

        if isinstance(raw, list):
            raise CannotParseAs(raw, self)

        parsed = parse(raw)
        if self.tz is not None:
            parsed = parsed.astimezone(self.tz)

        return parsed

    def __repr__(self) -> str:
        return "datetime"


@dataclass
class TimeDelta(ClypiParser[timedelta]):
    gt: timedelta | None = None
    gte: timedelta | None = None
    lt: timedelta | None = None
    lte: timedelta | None = None
    max: timedelta | None = None
    min: timedelta | None = None

    TIMEDELTA_UNITS = {
        ("weeks", "week", "w"): "weeks",
        ("days", "day", "d"): "days",
        ("hours", "hour", "h"): "hours",
        ("minutes", "minute", "m"): "minutes",
        ("seconds", "second", "s"): "seconds",
        ("milliseconds", "millisecond", "ms"): "milliseconds",
        ("microseconds", "microsecond", "us"): "microseconds",
    }
    TIMEDELTA_REGEX = re.compile(r"^(\d+)\s*(\w+)$")

    def __call__(self, raw: str | list[str], /) -> timedelta:
        if isinstance(raw, timedelta):
            return raw

        if not isinstance(raw, str):
            raise ValueError(
                f"Cannot parse {raw!r} as timedelta. Expected str or timedelta, got {type(raw).__name__}"
            )

        match = self.TIMEDELTA_REGEX.match(raw)
        if match is None:
            raise ValueError(f"Invalid timedelta {raw!r}.")

        value, unit = match.groups()
        parsed = None
        for units in self.TIMEDELTA_UNITS:
            if unit in units:
                parsed = timedelta(**{self.TIMEDELTA_UNITS[units]: int(value)})
                break

        if parsed is None:
            raise ValueError(f"Invalid timedelta {raw!r}.")

        if self.gt is not None:
            a(parsed > self.gt, parsed, f"is not greater than {self.gt}")
        if self.gte is not None:
            a(parsed >= self.gte, parsed, f"is not greater than or equal to {self.gte}")
        if self.lt is not None:
            a(parsed < self.lt, parsed, f"is not less than {self.lt}")
        if self.lte is not None:
            a(parsed <= self.lte, parsed, f"is not less than or equal to {self.lte}")
        if self.max is not None:
            a(parsed <= self.max, parsed, f"is not less than or equal to {self.max}")
        if self.min is not None:
            a(parsed >= self.min, parsed, f"is not greater than or equal to {self.min}")

        return parsed

    def __repr__(self) -> str:
        return "timedelta"


@dataclass
class Path(ClypiParser[_Path]):
    exists: bool = False

    def __call__(self, raw: str | list[str], /) -> _Path:
        if isinstance(raw, list):
            raise CannotParseAs(raw, self)
        p = _Path(raw)

        # Validations on the path
        if self.exists and not p.exists():
            raise ValueError(f"File {p.resolve()} does not exist!")

        return p

    def __repr__(self) -> str:
        return "path"


class List(ClypiParser[list[X]]):
    def __init__(self, inner: Parser[X]) -> None:
        self._inner = inner

    def __call__(self, raw: str | list[str], /) -> list[X]:
        if isinstance(raw, str):
            raw = raw.split(",")
        return [self._inner(item) for item in raw]

    def __repr__(self) -> str:
        return f"list({self._inner})"


class Tuple(ClypiParser[tuple[t.Any]]):
    def __init__(self, *inner: Parser[t.Any], num: int | None = None) -> None:
        self._inner = list(inner)
        self._num = num

    # TODO: can we return the right type here?
    def __call__(self, raw: str | list[str], /) -> tuple[t.Any, ...]:
        if isinstance(raw, str):
            raw = raw.split(",")

        if self._num and len(raw) != self._num:
            raise ValueError(
                f"Expected tuple of length {self._num} but instead got {len(raw)} items"
            )

        # Get all parsers for each item in the tuple (or reuse if tuple[T, ...])
        if not self._num:
            inner_parsers = [self._inner[0]] * len(raw)
        else:
            inner_parsers = self._inner

        # Parse each item with it's corresponding parser
        return tuple(parser(raw_item) for parser, raw_item in zip(inner_parsers, raw))

    def __repr__(self) -> str:
        args = ", ".join(str(it) for it in self._inner)
        return f"tuple({args})"


class Union(ClypiParser[t.Union[X, Y]]):
    def __init__(self, left: Parser[X], right: Parser[Y]) -> None:
        self._left = left
        self._right = right

    def __call__(self, raw: str | list[str], /) -> t.Union[X, Y]:
        # Str classes are catch-alls, so we de-prioritize them in unions
        # so that the other type is parsed first
        first, second = self._left, self._right
        if isinstance(first, Str):
            first, second = self._right, self._left

        with suppress(*CATCH_ERRORS):
            return first(raw)
        with suppress(*CATCH_ERRORS):
            return second(raw)
        raise CannotParseAs(raw, self)

    def _parts(self):
        """
        Some recursive magic here to "flatten" unions
        """
        parts: list[str] = []
        if left_parts := getattr(self._left, "_parts", None):
            parts.extend(left_parts())
        else:
            parts.append(str(self._left))

        if right_parts := getattr(self._right, "_parts", None):
            parts.extend(right_parts())
        else:
            parts.append(str(self._right))

        return parts

    def __repr__(self):
        return "(" + "|".join(self._parts()) + ")"


class Literal(ClypiParser[t.Any]):
    def __init__(self, *values: t.Any) -> None:
        self._values = values

    # TODO: can we return the right type here?
    def __call__(self, raw: str | list[str], /) -> t.Any:
        for v in self._values:
            if v == raw:
                return v
        raise CannotParseAs(raw, self)

    def __repr__(self):
        values = "|".join(str(v) for v in self._values)
        return "{" + values + "}"


class NoneParser(ClypiParser[None]):
    def __call__(self, raw: str | list[str], /) -> None:
        if not raw:
            return None
        if isinstance(raw, str) and raw.lower() == "none":
            return None
        raise CannotParseAs(raw, self)

    def __repr__(self):
        return "none"


class Enum(ClypiParser[type[enum.Enum]]):
    def __init__(self, _type: type[enum.Enum]) -> None:
        self._type = _type

    def __call__(self, raw: str | list[str], /) -> t.Any:
        if not isinstance(raw, str):
            raise CannotParseAs(raw, self)

        for enum_val in self._type:
            if raw.lower() == enum_val.name.lower():
                return enum_val

        raise ValueError(f"Value {raw} is not a valid choice between {self}")

    def __repr__(self):
        values = "|".join(v.name.lower() for v in self._type)
        return "{" + values + "}"


@tu.ignore_annotated
def from_type(_type: type) -> Parser[t.Any]:
    if _type is bool:
        return Bool()

    if _type is int:
        return Int()

    if _type is float:
        return Float()

    if _type is str:
        return Str()

    if _type is _Path:
        return Path()

    if _type is datetime:
        return DateTime()

    if _type is timedelta:
        return TimeDelta()

    if tu.is_list(_type):
        inner = from_type(tu.list_inner(_type))
        return List(inner)

    if tu.is_tuple(_type):
        inner_types, num = tu.tuple_inner(_type)
        inner_parsers = [from_type(it) for it in inner_types]
        return Tuple(*inner_parsers, num=num)

    if tu.is_union(_type):
        union_inner = tu.union_inner(_type)
        trav = Union(from_type(union_inner[0]), from_type(union_inner[1]))
        for rest in union_inner[2:]:
            trav = Union(trav, from_type(rest))
        return trav

    if tu.is_literal(_type):
        return Literal(*tu.literal_inner(_type))

    if tu.is_none(_type):
        return NoneParser()

    if tu.is_enum(_type):
        return Enum(_type)

    raise UnparseableException(f"Don't know how to parse as {_type}")



import typing as t
from dataclasses import asdict, dataclass

from clypi import _type_util
from clypi._cli import arg_parser
from clypi._exceptions import ClypiException
from clypi._prompts import MAX_ATTEMPTS
from clypi._util import UNSET, Unset
from clypi.parsers import Parser

T = t.TypeVar("T")

Nargs: t.TypeAlias = t.Literal["*", "+"] | float


@dataclass
class PartialConfig(t.Generic[T]):
    parser: Parser[T] | None = None
    default: T | Unset = UNSET
    default_factory: t.Callable[[], T] | Unset = UNSET
    help: str | None = None
    short: str | None = None
    prompt: str | None = None
    hide_input: bool = False
    max_attempts: int = MAX_ATTEMPTS
    inherited: bool = False
    hidden: bool = False
    group: str | None = None
    defer: bool = False


@dataclass
class Config(t.Generic[T]):
    name: str
    parser: Parser[T]
    arg_type: t.Any
    default: T | Unset = UNSET
    default_factory: t.Callable[[], T] | Unset = UNSET
    help: str | None = None
    short: str | None = None
    prompt: str | None = None
    hide_input: bool = False
    max_attempts: int = MAX_ATTEMPTS
    inherited: bool = False
    hidden: bool = False
    group: str | None = None
    defer: bool = False

    def __post_init__(self):
        if self.is_positional and self.short:
            raise ClypiException("Positional arguments cannot have short names")
        if self.is_positional and self.group:
            raise ClypiException("Positional arguments cannot belong to groups")

    def has_default(self) -> bool:
        return not isinstance(self.default, Unset) or not isinstance(
            self.default_factory, Unset
        )

    def get_default(self) -> T:
        val = self.get_default_or_missing()
        if isinstance(val, Unset):
            raise ValueError(f"Field {self} has no default value!")
        return val

    def get_default_or_missing(self) -> T | Unset:
        if not isinstance(self.default, Unset):
            return self.default
        if not isinstance(self.default_factory, Unset):
            return self.default_factory()
        return UNSET

    @classmethod
    def from_partial(
        cls,
        partial: PartialConfig[T],
        name: str,
        parser: Parser[T],
        arg_type: t.Any,
    ):
        kwargs = asdict(partial)
        kwargs.update(name=name, parser=parser, arg_type=arg_type)
        return cls(**kwargs)

    @property
    def display_name(self):
        name = arg_parser.snake_to_dash(self.name)
        if self.is_opt:
            return f"--{name}"
        return name

    @property
    def short_display_name(self):
        assert self.short, f"Expected short to be set in {self}"
        name = arg_parser.snake_to_dash(self.short)
        return f"-{name}"

    @property
    def is_positional(self) -> bool:
        if t.get_origin(self.arg_type) != t.Annotated:
            return False

        metadata = self.arg_type.__metadata__
        for m in metadata:
            if isinstance(m, _Positional):
                return True

        return False

    @property
    def is_opt(self) -> bool:
        return not self.is_positional

    @property
    def nargs(self) -> Nargs:
        if self.arg_type is bool:
            return 0

        if _type_util.is_list(self.arg_type):
            return "*"

        if _type_util.is_tuple(self.arg_type):
            sz = _type_util.tuple_size(self.arg_type)
            return "+" if sz == float("inf") else sz

        return 1

    @property
    def modifier(self) -> str:
        nargs = self.nargs
        if nargs in ("+", "*"):
            return "…"
        elif isinstance(nargs, int) and nargs > 1:
            return "…"
        return ""


def arg(
    default: T | Unset = UNSET,
    parser: Parser[T] | None = None,
    default_factory: t.Callable[[], T] | Unset = UNSET,
    help: str | None = None,
    short: str | None = None,
    prompt: str | None = None,
    hide_input: bool = False,
    max_attempts: int = MAX_ATTEMPTS,
    inherited: bool = False,
    hidden: bool = False,
    group: str | None = None,
    defer: bool = False,
) -> T:
    return PartialConfig(
        default=default,
        parser=parser,
        default_factory=default_factory,
        help=help,
        short=short,
        prompt=prompt,
        hide_input=hide_input,
        max_attempts=max_attempts,
        inherited=inherited,
        hidden=hidden,
        group=group,
        defer=defer,
    )  # type: ignore


@dataclass
class _Positional:
    pass


P = t.TypeVar("P")
Positional: t.TypeAlias = t.Annotated[P, _Positional()]



import pytest

from clypi import closest, distance


@pytest.mark.parametrize(
    "this,other,expected",
    [
        ("a", "a", 0),
        ("a", "A", 0.5),
        ("a", "b", 1),
        ("aa", "bb", 2),
        ("ab", "bb", 1),
        ("", "bb", 2),
        ("aa", "", 2),
        ("wrapped", "tapped", 2),
        ("that", "this", 2),
    ],
)
def test_distance(this: str, other: str, expected: str):
    assert distance(this, other) == expected
    assert distance(other, this) == expected


@pytest.mark.parametrize(
    "this,others,expected",
    [
        ("v", ["v", "version", "foo"], ("v", 0)),
        ("v", ["V", "version", "foo"], ("V", 0.5)),
        ("a", ["b", "version", "foo"], ("b", 1)),
        ("that", ["this", "foo"], ("this", 2)),
    ],
)
def test_closest(this: str, others: list[str], expected: tuple[str, int]):
    assert closest(this, others) == expected
